/*
 * kgraft_patch_bsc1021417
 *
 * Fix for bsc#1021417
 *
 *  SLE12-SP1 commit:
 *  d7c3c33c7c75 ("netback: correct array index (bsc#983348).")
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if defined(CONFIG_XEN)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/netdevice.h>
#include <xen/gnttab.h>
#include <xen/interface/io/ring.h>
#include <xen/barrier.h>
#include <xen/net-util.h>

#include "local.h"

#define KGR_PATCHED_MODULE      "netbk"

/* vmlinux kallsyms */
static void (*kgr_notify_remote_via_irq)(int irq);

/* netbk kallsyms */
static struct xen_netbk **__read_mostly kgr_xen_netbk;
static bool *kgr_MODPARM_permute_returns;
static unsigned int __read_mostly *kgr_netbk_nr_groups;
static int *kgr_netbk_copy_skb_mode;
static void (*kgr_flush_notify_list)(netif_t *list, unsigned int idx, multicall_entry_t mcl[], unsigned int limit);
static void (*kgr_remove_from_net_schedule_list)(netif_t *netif);
static void (*kgr_netbk_fatal_tx_err)(netif_t *netif);
static void (*kgr_tx_credit_callback)(unsigned long data);
static int (*kgr_netbk_get_extras)(netif_t *netif, struct netif_extra_info *extras, int work_to_do);
static void (*kgr_netbk_tx_err)(netif_t *netif, netif_tx_request_t *txp, RING_IDX end);
static unsigned int *kgr_max_tx_slots;
static void (*kgr_netif_schedule_work)(netif_t *netif);
static void (*kgr_netif_idx_release)(struct xen_netbk *netbk, u16 pending_idx);
static bool __read_mostly *kgr_use_kthreads;
static unsigned *kgr_random_src;

static struct kgr_kallsyms {
	char *name;
	void **addr;
} kgr_funcs_vmlinux[] = {
	{ "notify_remote_via_irq", (void *)&kgr_notify_remote_via_irq },
}, kgr_funcs_netbk[] = {
	{ "xen_netbk", (void *)&kgr_xen_netbk },
	{ "MODPARM_permute_returns", (void *)&kgr_MODPARM_permute_returns },
	{ "netbk_nr_groups", (void *)&kgr_netbk_nr_groups },
	{ "netbk_copy_skb_mode", (void *)&kgr_netbk_copy_skb_mode },
	{ "flush_notify_list", (void *)&kgr_flush_notify_list },
	{ "remove_from_net_schedule_list", (void *)&kgr_remove_from_net_schedule_list },
	{ "netbk_fatal_tx_err", (void *)&kgr_netbk_fatal_tx_err },
	{ "tx_credit_callback", (void *)&kgr_tx_credit_callback },
	{ "netbk_get_extras", (void *)&kgr_netbk_get_extras },
	{ "netbk_tx_err", (void *)&kgr_netbk_tx_err },
	{ "max_tx_slots", (void *)&kgr_max_tx_slots },
	{ "netif_schedule_work", (void *)&kgr_netif_schedule_work },
	{ "netif_idx_release", (void *)&kgr_netif_idx_release },
	{ "use_kthreads", (void *)&kgr_use_kthreads },
	{ "random_src.50871", (void *)&kgr_random_src },
};

/* inlined */
static u16 kgr_frag_get_pending_idx(const skb_frag_t *frag)
{
	return (u16)frag->page_offset;
}

/* isra optimized, the very patched function */
static void kgr_netbk_fill_tx_copy(const netif_tx_request_t *txreq,
			       struct netbk_tx_gop *gop, domid_t domid)
{
	gop->copy--;
	gop->copy->source.u.ref = txreq->gref;
	gop->copy->source.domid = domid;
	gop->copy->source.offset = txreq->offset;
	gop->copy->dest.u.gmfn = virt_to_mfn(gop->ptr);
	gop->copy->dest.domid = DOMID_SELF;
	gop->copy->dest.offset = offset_in_page(gop->ptr);
	gop->copy->flags = GNTCOPY_source_gref;

	if (gop->copy->dest.offset + txreq->size > PAGE_SIZE) {
		unsigned int first = PAGE_SIZE - gop->copy->dest.offset;

		gop->copy->len = first;
		gop->ptr += first;

		gop->copy--;
		/*
		 * Fix bsc#1021417
		 *  -gop->copy->source = gop->copy[-1].source;
		 *  +1 line
		 */
		gop->copy->source = gop->copy[1].source;
		gop->copy->source.offset += first;
		gop->copy->dest.u.gmfn = virt_to_mfn(gop->ptr);
		gop->copy->dest.domid = DOMID_SELF;
		gop->copy->dest.offset = 0;
		gop->copy->flags = GNTCOPY_source_gref;
		gop->copy->len = txreq->size - first;
	} else
		gop->copy->len = txreq->size;

	gop->ptr += gop->copy->len;
}

/* inlined */
static inline unsigned long kgr_idx_to_pfn(struct xen_netbk *netbk, u16 idx)
{
	return page_to_pfn(netbk->tx.mmap_pages[idx]);
}

/* inlined */
static inline unsigned long kgr_idx_to_kaddr(struct xen_netbk *netbk, u16 idx)
{
	return (unsigned long)pfn_to_kaddr(kgr_idx_to_pfn(netbk, idx));
}

/* inlined */
static void kgr_frag_set_pending_idx(skb_frag_t *frag, u16 pending_idx)
{
	frag->page_offset = pending_idx;
}

/* patched function */
void kgr_netbk_get_requests(struct xen_netbk *netbk, netif_t *netif,
	struct sk_buff *skb, struct netbk_tx_gop *gop)
{
	netif_tx_request_t *txp = netbk->tx.slots;
	struct pending_tx_info *pending_tx_info = netbk->tx.pending_info;
	struct skb_shared_info *shinfo = skb_shinfo(skb);
	skb_frag_t *frags = shinfo->frags;
	u16 pending_idx = netbk_tx_cb(skb)->pending_idx[0];
	pending_ring_idx_t index;
	int i, start;

	/* Skip first skb fragment if it is on same page as header fragment. */
	start = (kgr_frag_get_pending_idx(frags) == pending_idx);

	for (i = 0; i < netbk_tx_cb(skb)->copy_slots; ++i, txp++) {
		index = MASK_PEND_IDX(netbk->tx.pending_cons++);
		pending_idx = netbk->tx.pending_ring[index];

		kgr_netbk_fill_tx_copy(txp, gop, netif->domid);

		pending_tx_info[pending_idx].req = *txp;
		netif_get(netif);
		pending_tx_info[pending_idx].netif = netif;
		netbk_tx_cb(skb)->pending_idx[1 + i] = pending_idx;
	}

	for (i = start; i < shinfo->nr_frags; i++, txp++) {
		index = MASK_PEND_IDX(netbk->tx.pending_cons++);
		pending_idx = netbk->tx.pending_ring[index];

		gnttab_set_map_op(gop->map++, kgr_idx_to_kaddr(netbk, pending_idx),
				  GNTMAP_host_map | GNTMAP_readonly,
				  txp->gref, netif->domid);

		memcpy(&pending_tx_info[pending_idx].req, txp, sizeof(*txp));
		netif_get(netif);
		pending_tx_info[pending_idx].netif = netif;
		kgr_frag_set_pending_idx(&frags[i], pending_idx);
	}

	if ((void *)gop->map > (void *)gop->copy)
		net_warn_ratelimited("%s: Grant op overrun (%p > %p)\n",
				     netdev_name(netif->dev),
				     gop->map, gop->copy);
}

/* inlined */
static void kgr_permute_dealloc_ring(u16 *dealloc_ring, pending_ring_idx_t dc,
				 pending_ring_idx_t dp)
{
	/*
	 * Fix bsc#1021417
	 *  -static unsigned random_src = 0x12345678;
	 */
	unsigned dst_offset;
	pending_ring_idx_t dest;
	u16 tmp;

	while (dc != dp) {
		dst_offset = (*kgr_random_src / 256) % (dp - dc);
		dest = dc + dst_offset;
		tmp = dealloc_ring[MASK_PEND_IDX(dest)];
		dealloc_ring[MASK_PEND_IDX(dest)] =
			dealloc_ring[MASK_PEND_IDX(dc)];
		dealloc_ring[MASK_PEND_IDX(dc)] = tmp;
		dc++;
		*kgr_random_src *= 68389;
	}
}

/* inlined */
static inline int kgr_copy_pending_req(struct xen_netbk *netbk,
				   pending_ring_idx_t pending_idx)
{
	return gnttab_copy_grant_page(netbk->tx.pending_info[pending_idx].grant_handle,
				      &netbk->tx.mmap_pages[pending_idx]);
}

/* inlined */
static bool kgr_make_tx_response(netif_t *netif, const netif_tx_request_t *txp,
			     s8 st, netif_t **tailp)
{
	RING_IDX i = netif->tx.rsp_prod_pvt;
	netif_tx_response_t *resp;
	int notify;

	resp = RING_GET_RESPONSE(&netif->tx, i);
	resp->id     = txp->id;
	resp->status = st;

	if (txp->flags & XEN_NETTXF_extra_info)
		RING_GET_RESPONSE(&netif->tx, ++i)->status = XEN_NETIF_RSP_NULL;

	netif->tx.rsp_prod_pvt = ++i;
	RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&netif->tx, notify);
	if (notify) {
		if (!tailp)
			kgr_notify_remote_via_irq(netif->irq);
		else if (netif != *tailp && !netif->tx_notify_link) {
			if (*tailp)
				(*tailp)->tx_notify_link = netif;
			*tailp = netif;
		} else
			notify = false;
	}

#ifdef CONFIG_XEN_NETDEV_PIPELINED_TRANSMITTER
	if (i == netif->tx.req_cons) {
		int more_to_do;
		RING_FINAL_CHECK_FOR_REQUESTS(&netif->tx, more_to_do);
		if (more_to_do)
			add_to_net_schedule_list_tail(netif);
	}
#endif

	return notify;
}

/* inlined */
inline static void kgr_net_tx_action_dealloc(struct xen_netbk *netbk)
{
	struct pending_tx_info *pending_tx_info = netbk->tx.pending_info;
	struct netbk_tx_pending_inuse *inuse, *n;
	gnttab_unmap_grant_ref_t *gop;
	u16 pending_idx;
	pending_ring_idx_t dc, dp;
	netif_t *netif, *notify_head = NULL, *notify_tail = NULL;
	LIST_HEAD(list);

	dc = netbk->tx.dealloc_cons;
	gop = netbk->tx.unmap_ops;

	/*
	 * Free up any grants we have finished using
	 */
	do {
		dp = netbk->tx.dealloc_prod;

		/* Ensure we see all indices enqueued by netif_idx_release(). */
		smp_rmb();

		if (*kgr_MODPARM_permute_returns && *kgr_netbk_nr_groups == 1)
			kgr_permute_dealloc_ring(netbk->tx.dealloc_ring, dc, dp);

		while (dc != dp) {
			unsigned long pfn;
			struct netbk_tx_pending_inuse *pending_inuse =
					netbk->tx.pending_inuse;

			pending_idx = netbk->tx.dealloc_ring[MASK_PEND_IDX(dc++)];
			list_move_tail(&pending_inuse[pending_idx].list, &list);

			pfn = kgr_idx_to_pfn(netbk, pending_idx);
			/* Already unmapped? */
			if (!phys_to_machine_mapping_valid(pfn))
				continue;

			gnttab_set_unmap_op(gop, kgr_idx_to_kaddr(netbk, pending_idx),
					    GNTMAP_host_map,
					    netbk->tx.pending_info[pending_idx].grant_handle);
			gop++;
		}

	} while (dp != netbk->tx.dealloc_prod);

	netbk->tx.dealloc_cons = dc;

	if (HYPERVISOR_grant_table_op(GNTTABOP_unmap_grant_ref,
				      netbk->tx.unmap_ops,
				      gop - netbk->tx.unmap_ops))
		BUG();

	/* Copy any entries that have been pending for too long. */
	if (*kgr_netbk_copy_skb_mode == NETBK_DELAYED_COPY_SKB &&
	    !list_empty(&netbk->tx.pending_inuse_head)) {
		list_for_each_entry_safe(inuse, n, &netbk->tx.pending_inuse_head, list) {
			if (time_after(inuse->alloc_time + HZ / 2, jiffies))
				break;

			pending_idx = inuse - netbk->tx.pending_inuse;

			pending_tx_info[pending_idx].netif->nr_copied_tx_skbs++;

			switch (kgr_copy_pending_req(netbk, pending_idx)) {
			case 0:
				list_move_tail(&inuse->list, &list);
				continue;
			case -EBUSY:
				list_del_init(&inuse->list);
				continue;
			case -ENOENT:
				continue;
			}

			break;
		}
	}

	list_for_each_entry_safe(inuse, n, &list, list) {
		pending_idx = inuse - netbk->tx.pending_inuse;
		netif = pending_tx_info[pending_idx].netif;

		if (!kgr_make_tx_response(netif, &pending_tx_info[pending_idx].req,
				      XEN_NETIF_RSP_OKAY, &notify_tail))
			netif_put(netif);
		else if (!notify_head)
			notify_head = netif;

		/* Ready for next use. */
		gnttab_reset_grant_page(netbk->tx.mmap_pages[pending_idx]);

		netbk->tx.pending_ring[MASK_PEND_IDX(netbk->tx.pending_prod++)] =
			pending_idx;

		list_del_init(&inuse->list);
	}
	if (notify_head)
		kgr_flush_notify_list(notify_head, TX_IDX, netbk->tx.mcl,
				  sizeof(netbk->tx.map_ops)
				  / sizeof(*netbk->tx.mcl));
}

/* inlined */
static inline pending_ring_idx_t kgr_nr_pending_reqs(const struct xen_netbk *netbk)
{
	return MAX_PENDING_REQS -
		netbk->tx.pending_prod + netbk->tx.pending_cons;
}

/* inlined */
static netif_t *kgr_poll_net_schedule_list(struct xen_netbk *netbk)
{
	netif_t *netif = NULL;

	spin_lock_irq(&netbk->tx.schedule_list_lock);
	if (!list_empty(&netbk->tx.schedule_list)) {
		netif = list_first_entry(&netbk->tx.schedule_list,
					 netif_t, list);
		netif_get(netif);
		kgr_remove_from_net_schedule_list(netif);
	}
	spin_unlock_irq(&netbk->tx.schedule_list_lock);
	return netif;
}

/* inlined */
static void kgr_tx_add_credit(netif_t *netif)
{
	unsigned long max_burst, max_credit;

	/*
	 * Allow a burst big enough to transmit a jumbo packet of up to 128kB.
	 * Otherwise the interface can seize up due to insufficient credit.
	 */
	max_burst = max(131072UL, netif->credit_bytes);

	/* Take care that adding a new chunk of credit doesn't wrap to zero. */
	max_credit = netif->remaining_credit + netif->credit_bytes;
	if (max_credit < netif->remaining_credit)
		max_credit = ULONG_MAX; /* wrapped: clamp to ULONG_MAX */

	netif->remaining_credit = min(max_credit, max_burst);
}

/* inlined */
static int kgr_netbk_count_requests(netif_t *netif, netif_tx_request_t *first,
				netif_tx_request_t *txp, int work_to_do)
{
	RING_IDX cons = netif->tx.req_cons;
	int slots = 0, drop_err = 0;

	if (!(first->flags & XEN_NETTXF_more_data))
		return 0;

	do {
		if (slots >= work_to_do) {
			netdev_err(netif->dev, "Need more slots\n");
			kgr_netbk_fatal_tx_err(netif);
			return -ENODATA;
		}

		if (unlikely(slots >= *kgr_max_tx_slots)) {
			netdev_err(netif->dev, "Too many slots\n");
			kgr_netbk_fatal_tx_err(netif);
			return -E2BIG;
		}

		/*
		 * The Xen network protocol had an implicit dependency on
		 * MAX_SKB_FRAGS. XEN_NETIF_NR_SLOTS_MIN is set to the
		 * historical MAX_SKB_FRAGS value 18 to honor the same
		 * behavior as before. Any packet using more than 18 slots
		 * but less than max_tx_slots slots is dropped.
		 */
		switch (slots) {
		case 0 ... XEN_NETIF_NR_SLOTS_MIN - 1:
			break;
		case XEN_NETIF_NR_SLOTS_MIN:
			if (net_ratelimit())
				netdev_dbg(netif->dev,
					   "slot count exceeding limit of %d, dropping packet\n",
					   XEN_NETIF_NR_SLOTS_MIN);
			if (!drop_err)
				drop_err = -E2BIG;
			/* fall through */
		default:
			--txp;
			break;
		}

		*txp = *RING_GET_REQUEST(&netif->tx, cons + slots);
		barrier();

		/*
		 * If the guest submitted a frame >= 64 KiB then first->size
		 * overflowed and following slots will appear to be larger
		 * than the frame. This cannot be fatal error as there are
		 * buggy frontends that do this.
		 *
		 * Consume all slots and drop the packet.
		 */
		if (!drop_err && txp->size > first->size) {
			if (net_ratelimit())
				netdev_dbg(netif->dev,
					   "Invalid tx request (slot size %u > remaining size %u)\n",
					   txp->size, first->size);
			drop_err = -EIO;
		}

		first->size -= txp->size;
		slots++;

		if (unlikely((txp->offset + txp->size) > PAGE_SIZE)) {
			netdev_err(netif->dev, "txp->offset: %x, size: %u\n",
				   txp->offset, txp->size);
			kgr_netbk_fatal_tx_err(netif);
			return -EINVAL;
		}
	} while ((txp++)->flags & XEN_NETTXF_more_data);

	if (drop_err) {
		kgr_netbk_tx_err(netif, first, cons + slots);
		return drop_err;
	}

	return slots;
}

/* inlined */
static int kgr_netbk_set_skb_gso(netif_t *netif, struct sk_buff *skb,
			     struct netif_extra_info *gso)
{
	if (!gso->u.gso.size) {
		netdev_err(skb->dev, "GSO size must not be zero.\n");
		kgr_netbk_fatal_tx_err(netif);
		return -EINVAL;
	}

	/* Currently only TCPv4 S.O. is supported. */
	if (gso->u.gso.type != XEN_NETIF_GSO_TYPE_TCPV4) {
		netdev_err(skb->dev, "Bad GSO type %d.\n", gso->u.gso.type);
		kgr_netbk_fatal_tx_err(netif);
		return -EINVAL;
	}

	skb_shinfo(skb)->gso_size = gso->u.gso.size;
	skb_shinfo(skb)->gso_type = SKB_GSO_TCPV4;

	/* Header must be checked, and gso_segs computed. */
	skb_shinfo(skb)->gso_type |= SKB_GSO_DODGY;
	skb_shinfo(skb)->gso_segs = 0;

	return 0;
}

/* inlined */
static int kgr_netbk_tx_check_gop(struct xen_netbk *netbk, struct sk_buff *skb,
			      struct netbk_tx_gop *gop, bool hdr_copied)
{
	gnttab_copy_t *cop = gop->copy;
	gnttab_map_grant_ref_t *mop = gop->map;
	u16 pending_idx = netbk_tx_cb(skb)->pending_idx[0];
	pending_ring_idx_t index;
	struct pending_tx_info *pending_tx_info = netbk->tx.pending_info;
	netif_t *netif = pending_tx_info[pending_idx].netif;
	netif_tx_request_t *txp = &pending_tx_info[pending_idx].req;
	struct skb_shared_info *shinfo = skb_shinfo(skb);
	int nr_frags = shinfo->nr_frags;
	int i, err, start;

	/* Check status of header. */
	if (hdr_copied) {
		err = (--cop)->status;
		if (txp->size > cop->len)
			cmpxchg_local(&err, GNTST_okay, (--cop)->status);
		if (!kgr_make_tx_response(netif, txp,
				      err == GNTST_okay ? XEN_NETIF_RSP_OKAY
							: XEN_NETIF_RSP_ERROR,
				      &gop->notify.tail))
			netif_put(netif);
		else if (!gop->notify.head)
			gop->notify.head = netif;
		index = MASK_PEND_IDX(netbk->tx.pending_prod++);
		netbk->tx.pending_ring[index] = pending_idx;
	} else if (unlikely((err = mop->status) != GNTST_okay)) {
		++mop;
		kgr_make_tx_response(netif, txp, XEN_NETIF_RSP_ERROR, NULL);
		index = MASK_PEND_IDX(netbk->tx.pending_prod++);
		netbk->tx.pending_ring[index] = pending_idx;
		netif_put(netif);
	} else {
		set_phys_to_machine(kgr_idx_to_pfn(netbk, pending_idx),
			FOREIGN_FRAME(mop->dev_bus_addr >> PAGE_SHIFT));
		container_of(txp, struct pending_tx_info, req)->grant_handle
			= mop++->handle;
	}

	/* Skip first skb fragment if it is on same page as header fragment. */
	start = (kgr_frag_get_pending_idx(shinfo->frags) == pending_idx);

	for (i = 0; i < netbk_tx_cb(skb)->copy_slots; ++i) {
		int newerr = (--cop)->status;

		pending_idx = netbk_tx_cb(skb)->pending_idx[1 + i];
		txp = &pending_tx_info[pending_idx].req;
		if (txp->size > cop->len)
			cmpxchg_local(&newerr, GNTST_okay, (--cop)->status);
		if (!kgr_make_tx_response(netif, txp,
				      newerr == GNTST_okay ? XEN_NETIF_RSP_OKAY
							   : XEN_NETIF_RSP_ERROR,
				      &gop->notify.tail))
			netif_put(netif);
		else if (!gop->notify.head)
			gop->notify.head = netif;
		cmpxchg_local(&err, GNTST_okay, newerr);
		index = MASK_PEND_IDX(netbk->tx.pending_prod++);
		netbk->tx.pending_ring[index] = pending_idx;
	}

	for (i = start; i < nr_frags; i++, mop++) {
		int j, newerr;

		pending_idx = kgr_frag_get_pending_idx(&shinfo->frags[i]);
		txp = &pending_tx_info[pending_idx].req;

		/* Check error status: if okay then remember grant handle. */
		newerr = mop->status;
		if (likely(newerr == GNTST_okay)) {
			set_phys_to_machine(kgr_idx_to_pfn(netbk, pending_idx),
				FOREIGN_FRAME(mop->dev_bus_addr>>PAGE_SHIFT));
			container_of(txp, struct pending_tx_info, req)->grant_handle
				= mop->handle;
			/* Had a previous error? Invalidate this fragment. */
			if (unlikely(err != GNTST_okay))
				kgr_netif_idx_release(netbk, pending_idx);
			continue;
		}

		/* Error on this fragment: respond to client with an error. */
		kgr_make_tx_response(netif, txp, XEN_NETIF_RSP_ERROR, NULL);
		index = MASK_PEND_IDX(netbk->tx.pending_prod++);
		netbk->tx.pending_ring[index] = pending_idx;
		netif_put(netif);

		/* Not the first error? Preceding frags already invalidated. */
		if (err != GNTST_okay)
			continue;

		/* First error: invalidate header and preceding fragments. */
		if (!hdr_copied) {
			pending_idx = netbk_tx_cb(skb)->pending_idx[0];
			kgr_netif_idx_release(netbk, pending_idx);
		}
		for (j = start; j < i; j++) {
			pending_idx = kgr_frag_get_pending_idx(&shinfo->frags[j]);
			kgr_netif_idx_release(netbk, pending_idx);
		}

		/* Remember the error: invalidate all subsequent fragments. */
		err = newerr;
	}

	gop->map = mop;
	gop->copy = cop;
	if ((void *)mop > (void *)cop)
		net_warn_ratelimited("%s: Grant op check overrun (%p > %p)\n",
				     netdev_name(netif->dev), mop, cop);
	return err;
}

/* inlined */
static void kgr_netbk_fill_frags(struct xen_netbk *netbk, struct sk_buff *skb)
{
	struct skb_shared_info *shinfo = skb_shinfo(skb);
	int nr_frags = shinfo->nr_frags;
	int i;

	for (i = 0; i < nr_frags; i++) {
		netif_tx_request_t *txp;
		u16 pending_idx = kgr_frag_get_pending_idx(shinfo->frags + i);

		netbk->tx.pending_inuse[pending_idx].alloc_time = jiffies;
		list_add_tail(&netbk->tx.pending_inuse[pending_idx].list,
			      &netbk->tx.pending_inuse_head);

		txp = &netbk->tx.pending_info[pending_idx].req;
		__skb_fill_page_desc(skb, i, netbk->tx.mmap_pages[pending_idx],
				     txp->offset, txp->size);

		skb->len += txp->size;
		skb->data_len += txp->size;
		skb->truesize += txp->size;
	}
}

/* patched function */
void kgr_net_tx_action(unsigned long group)
{
	struct xen_netbk *netbk = &(*kgr_xen_netbk)[group];
	struct sk_buff *skb;
	netif_t *netif;
	netif_tx_request_t txreq, *txslot;
	struct netif_extra_info extras[XEN_NETIF_EXTRA_TYPE_MAX - 1];
	u16 pending_idx;
	RING_IDX i;
	struct netbk_tx_gop gop;
	multicall_entry_t mcl[2];
	unsigned int data_len;
	int ret, work_to_do;

	BUILD_BUG_ON(sizeof(skb->cb) < sizeof(struct netbk_tx_cb));

	kgr_net_tx_action_dealloc(netbk);

	gop.map = netbk->tx.map_ops;
	gop.copy = &netbk->tx.copy_op + 1;
	while (kgr_nr_pending_reqs(netbk) + XEN_NETIF_NR_SLOTS_MIN < MAX_PENDING_REQS
	       && !list_empty(&netbk->tx.schedule_list)) {
		/* Get a netif from the list with work to do. */
		netif = kgr_poll_net_schedule_list(netbk);
		/*
		 * This can sometimes happen because the test of
		 * list_empty(net_schedule_list) at the top of the
		 * loop is unlocked.  Just go back and have another
		 * look.
		 */
		if (!netif)
			continue;

		if (unlikely(netif->busted)) {
			netif_put(netif);
			continue;
		}

		if (netif->tx.sring->req_prod - netif->tx.req_cons >
		    NET_TX_RING_SIZE) {
			netdev_err(netif->dev,
				   "Impossible number of requests"
				   " (prod=%u cons=%u size=%lu)\n",
				   netif->tx.sring->req_prod,
				   netif->tx.req_cons, NET_TX_RING_SIZE);
			kgr_netbk_fatal_tx_err(netif);
			continue;
		}

		RING_FINAL_CHECK_FOR_REQUESTS(&netif->tx, work_to_do);
		if (!work_to_do) {
			netif_put(netif);
			continue;
		}

		i = netif->tx.req_cons;
		rmb(); /* Ensure that we see the request before we copy it. */
		memcpy(&txreq, RING_GET_REQUEST(&netif->tx, i), sizeof(txreq));
		barrier();

		/* Credit-based scheduling. */
		if (txreq.size > netif->remaining_credit) {
			u64 now = get_jiffies_64();
			u64 next_credit = netif->credit_window_start +
				msecs_to_jiffies(netif->credit_usec / 1000);

			/* Timer could already be pending in rare cases. */
			if (timer_pending(&netif->credit_timeout)) {
				netif_put(netif);
				continue;
			}

			/* Passed the point where we can replenish credit? */
			if (time_after_eq64(now, next_credit)) {
				netif->credit_window_start = now;
				kgr_tx_add_credit(netif);
			}

			/* Still too big to send right now? Set a callback. */
			if (txreq.size > netif->remaining_credit) {
				netif->credit_timeout.data     =
					(unsigned long)netif;
				netif->credit_timeout.function =
					kgr_tx_credit_callback;
				mod_timer(&netif->credit_timeout, next_credit);
				netif->credit_window_start = next_credit;
				netif_put(netif);
				continue;
			}
		}
		netif->remaining_credit -= txreq.size;

		work_to_do--;
		netif->tx.req_cons = ++i;

		memset(extras, 0, sizeof(extras));
		if (txreq.flags & XEN_NETTXF_extra_info) {
			work_to_do = kgr_netbk_get_extras(netif, extras,
						      work_to_do);
			i = netif->tx.req_cons;
			if (unlikely(work_to_do < 0))
				continue;
		}

		txslot = netbk->tx.slots;
		ret = kgr_netbk_count_requests(netif, &txreq, txslot, work_to_do);
		if (unlikely(ret < 0))
			continue;

		i += ret;

		if (unlikely(txreq.size < ETH_HLEN)) {
			netdev_dbg(netif->dev, "Bad packet size: %d\n",
				   txreq.size);
			kgr_netbk_tx_err(netif, &txreq, i);
			continue;
		}

		/* No crossing a page as the payload mustn't fragment. */
		if (unlikely((txreq.offset + txreq.size) > PAGE_SIZE)) {
			netdev_err(netif->dev,
				   "txreq.offset: %x, size: %u, end: %lu\n",
				   txreq.offset, txreq.size,
				   (txreq.offset & ~PAGE_MASK) + txreq.size);
			kgr_netbk_fatal_tx_err(netif);
			continue;
		}

		pending_idx = netbk->tx.pending_ring[MASK_PEND_IDX(netbk->tx.pending_cons)];

		data_len = (txreq.size > PKT_PROT_LEN &&
			    ret < MAX_SKB_FRAGS) ?
			PKT_PROT_LEN : txreq.size;
		while (ret > MAX_SKB_FRAGS ||
		       (ret && (data_len + txslot->size <= PKT_PROT_LEN ||
				*kgr_netbk_copy_skb_mode == NETBK_ALWAYS_COPY_SKB))) {
			data_len += txslot++->size;
			--ret;
		}

		skb = alloc_skb(data_len + 16 + NET_IP_ALIGN,
				GFP_ATOMIC | __GFP_NOWARN);
		if (unlikely(skb == NULL)) {
			netdev_dbg(netif->dev,
				   "Can't allocate a skb in start_xmit.\n");
			kgr_netbk_tx_err(netif, &txreq, i);
			break;
		}

		/* Packets passed to netif_rx() must have some headroom. */
		skb_reserve(skb, 16 + NET_IP_ALIGN);
		skb->dev = netif->dev;

		if (extras[XEN_NETIF_EXTRA_TYPE_GSO - 1].type) {
			struct netif_extra_info *gso;
			gso = &extras[XEN_NETIF_EXTRA_TYPE_GSO - 1];

			if (kgr_netbk_set_skb_gso(netif, skb, gso)) {
				/* Failure in netbk_set_skb_gso is fatal. */
				kfree_skb(skb);
				continue;
			}
		}

		netbk->tx.pending_info[pending_idx].req = txreq;
		netbk->tx.pending_info[pending_idx].netif = netif;
		netbk_tx_cb(skb)->pending_idx[0] = pending_idx;
		netbk_tx_cb(skb)->copy_slots = txslot - netbk->tx.slots;

		__skb_put(skb, data_len);
		gop.ptr = skb->data;

		skb_shinfo(skb)->nr_frags = ret;
		if (data_len < txreq.size) {
			gnttab_set_map_op(gop.map++,
					  kgr_idx_to_kaddr(netbk, pending_idx),
					  GNTMAP_host_map | GNTMAP_readonly,
					  txreq.gref, netif->domid);
			skb_shinfo(skb)->nr_frags++;
		} else {
			kgr_netbk_fill_tx_copy(&txreq, &gop, netif->domid);
			pending_idx = INVALID_PENDING_IDX;
		}
		kgr_frag_set_pending_idx(skb_shinfo(skb)->frags, pending_idx);

		__skb_queue_tail(&netbk->tx.queue, skb);

		netbk->tx.pending_cons++;

		kgr_netbk_get_requests(netbk, netif, skb, &gop);

		netif->tx.req_cons = i;
		kgr_netif_schedule_work(netif);
	}

	if (skb_queue_empty(&netbk->tx.queue))
		goto out;

    /* NOTE: some maps may fail with GNTST_eagain, which could be successfully
     * retried in the backend after a delay. However, we can also fail the tx
     * req and let the frontend resend the relevant packet again. This is fine
     * because it is unlikely that a network buffer will be paged out or shared,
     * and therefore it is unlikely to fail with GNTST_eagain. */
	MULTI_grant_table_op(&mcl[0], GNTTABOP_copy, gop.copy,
			     &netbk->tx.copy_op + 1 - gop.copy);
	MULTI_grant_table_op(&mcl[1], GNTTABOP_map_grant_ref,
			     netbk->tx.map_ops, gop.map - netbk->tx.map_ops);
	if (HYPERVISOR_multicall_check(mcl, 2, NULL))
		BUG();

	gop.map = netbk->tx.map_ops;
	gop.copy = &netbk->tx.copy_op + 1;
	gop.notify.head = NULL;
	gop.notify.tail = NULL;
	while ((skb = __skb_dequeue(&netbk->tx.queue)) != NULL) {
		struct net_device *dev;
		netif_tx_request_t *txp;

		pending_idx = netbk_tx_cb(skb)->pending_idx[0];
		netif       = netbk->tx.pending_info[pending_idx].netif;
		dev         = netif->dev;
		txp         = &netbk->tx.pending_info[pending_idx].req;
		data_len    = skb->len;

		/* Check the remap/copy error code. */
		if (unlikely(kgr_netbk_tx_check_gop(netbk, skb, &gop,
						data_len >= txp->size))) {
			netdev_dbg(dev, "netback grant failed.\n");
			skb_shinfo(skb)->nr_frags = 0;
			kfree_skb(skb);
			dev->stats.rx_errors++;
			continue;
		}

		if (data_len < txp->size) {
			memcpy(skb->data,
			       (void *)(kgr_idx_to_kaddr(netbk, pending_idx)
					+ txp->offset),
			       data_len);
			/* Append the packet payload as a fragment. */
			txp->offset += data_len;
			txp->size -= data_len;
		}

		if (txp->flags & XEN_NETTXF_csum_blank)
			skb->ip_summed = CHECKSUM_PARTIAL;
		else if (txp->flags & XEN_NETTXF_data_validated)
			skb->ip_summed = CHECKSUM_UNNECESSARY;
		else
			skb->ip_summed = CHECKSUM_NONE;

		kgr_netbk_fill_frags(netbk, skb);

		/*
		 * If the initial fragment was < PKT_PROT_LEN then
		 * pull through some bytes from the other fragments to
		 * increase the linear region to PKT_PROT_LEN bytes.
		 */
		if (skb_headlen(skb) < PKT_PROT_LEN && skb_is_nonlinear(skb)) {
			int target = min_t(int, skb->len, PKT_PROT_LEN);
			__pskb_pull_tail(skb, target - skb_headlen(skb));
		}

		skb->protocol = eth_type_trans(skb, dev);

		if (skb_checksum_setup(skb, &netif->rx_gso_csum_fixups)) {
			netdev_dbg(dev,
				   "Can't setup checksum in net_tx_action\n");
			kfree_skb(skb);
			dev->stats.rx_dropped++;
			continue;
		}

		skb_probe_transport_header(skb, 0);

		dev->stats.rx_bytes += skb->len;
		dev->stats.rx_packets++;

		if (*kgr_use_kthreads)
			netif_rx_ni(skb);
		else
			netif_rx(skb);
	}

	if (gop.notify.head)
		kgr_flush_notify_list(gop.notify.head, TX_IDX, netbk->tx.mcl,
				  sizeof(netbk->tx.map_ops)
				  / sizeof(*netbk->tx.mcl));

 out:
	if (*kgr_netbk_copy_skb_mode == NETBK_DELAYED_COPY_SKB &&
	    !list_empty(&netbk->tx.pending_inuse_head)) {
		struct netbk_tx_pending_inuse *oldest;

		oldest = list_entry(netbk->tx.pending_inuse_head.next,
				    struct netbk_tx_pending_inuse, list);
		mod_timer(&netbk->tx.pending_timer, oldest->alloc_time + HZ);
	}
}

static int kgr_patch_bsc1021417_kallsyms(struct kgr_kallsyms *kgr_funcs,
		int length)
{
	unsigned long addr;
	int i;

	for (i = 0; i < length; i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1021417_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1021417_kallsyms(kgr_funcs_netbk, ARRAY_SIZE(kgr_funcs_netbk));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1021417_module_nb = {
	.notifier_call = kgr_patch_bsc1021417_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1021417_init(void)
{
	int ret;

        ret = kgr_patch_bsc1021417_kallsyms(kgr_funcs_vmlinux,
                ARRAY_SIZE(kgr_funcs_vmlinux));
        if (ret)
                return ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1021417_kallsyms(kgr_funcs_netbk,
				ARRAY_SIZE(kgr_funcs_netbk));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1021417_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1021417_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1021417_module_nb);
}

#endif /* CONFIG_XEN */
